/*
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 */

#include <os_type.h>
#include <osapi.h>
#include <eagle_soc.h>
#include <ets_sys.h>

#include "supla_pow.h"
#include "supla_esp_gpio.h"
#include "supla_w1.h"
#include "supla-dev/log.h"
#include "pgmspace.h"
#include "gpio.h"
#include "user_config.h"
#include "supla_esp_devconn.h"

#ifdef POWSENSOR

//Interrupt Modes
#define RISING    0x01
#define FALLING   0x02
#define CHANGE    0x03
#define ONLOW     0x04
#define ONHIGH    0x05
#define ONLOW_WE  0x0C
#define ONHIGH_WE 0x0D

#define HIGH 0x1
#define LOW 0x0
#define INPUT 0x0
#define OUTPUT 0x1
#define INPUT_PULLUP 0x2
#define NOT_A_PIN 0
#define NOT_A_PORT 0
#define NOT_AN_INTERRUPT -1

extern uint32_t microseconds;
int supla_w13_pin;
int supla_w14_pin;
void DEVCONN_ICACHE_FLASH supla_micros();
void ICACHE_FLASH_ATTR sel_pin_current();
void ICACHE_FLASH_ATTR sel_pin_voltage();

// This is the factor of a voltage divider of 6x 470K upstream and 1k downstream
// as per recomended circuit in datasheet
#define R_VOLTAGE           2821

// Frequency of the HLW8012 internal clock
#define F_OSC               3579000

// Minimum delay between selecting a mode and reading a sample
#define READING_INTERVAL    3000

// Maximum pulse with in microseconds
// If longer than this pulse width is reset to 0
// This value is purely experimental.
// Higher values allow for a better precission but reduce sampling rate
// and response speed to change
// Lower values increase sampling rate but reduce precission
// Values below 0.5s are not recommended since current and voltage output
// will have no time to stabilise
#define PULSE_TIMEOUT       2000000

#define CALIBRATE_RUN 0       // do calibration to determine the inital values    enable = 1, disable = 0
                              // you can also switch to calibration mode when pressing the button while connecting to Wifi
// GPIOs
#define SEL_PIN    	   5	// GPIO5 GPIO5	 pin24
#define CF1_PIN        12	// MTCK  GPIO13  pin12
#define CF_PIN         9	// MTMS  GPIO14  pin9

// GPIO_ID_PIN(13) - dioda czerwona  / PWM0 / MTDI / GPIO12
// GPIO_ID_PIN(15) - LED_RED_PORT	 / PWM2 / MTDO / GPIO15  dioda niebieska

typedef struct {

    uint32_t _voltage;
    volatile unsigned char _mode;

    //bool _use_interrupts;
    volatile unsigned long _last_cf_interrupt;
    volatile unsigned long _last_cf1_interrupt;
    volatile unsigned long _first_cf1_interrupt;

    //void (*_checkCFSignal)();
    void (*_checkCF1Signal)();
    void (*_calculateDefaultMultipliers)();

}HLW8012;

HLW8012  hlw8012;			  

int run_cal =  CALIBRATE_RUN;
unsigned long lastmillis;
int counter;
unsigned long current_multiplier; 		 // Unit: us/A
unsigned int  voltage_multiplier;	 	 // Unit: us/V
unsigned long power_multiplier;   		 // Unit: us/W
int test_sel = 0;
uint32_t microseconds;
uint32_t voltage_time;
int supla_micros_gpio_state = 0;

// * cf_pin, cf1_pin and sel_pin are GPIOs to the HLW8012 IC
// * currentWhen is the value in sel_pin to select current sampling
// * set use_interrupts to true to use interrupts to monitor pulse widths
// * leave pulse_timeout to the default value, recommended when using interrupts
unsigned char currentWhen = HIGH;
unsigned long pulse_timeout = PULSE_TIMEOUT;

void noInterrupts();
void interrupts();

int stan = 0;			// odczyt stanu linii GPIO13
int od_ready = 0;
int stan_poprzedni = 0;	// zapamitany zmierzony stan impulsu przedostatniego pomiaru
int pomiar = 0;			// jesli = 1 to zacznij mierzy (przerwania supla_micros)
int sel_nap = 0;		// licznik 1-sekundowych pomiarw napicia
int licznik_przel = 6;	// okres pomiaru napicia / prdu
int licznik_pomiarow;	// obliczenie redniej ze 100 pomiarw
int nap_prad = 0;		// 0 - pomiar napicia, 1 - pomiar prdu

//-------------------------------------------------------
// Timer 5 mikrosekund
//-------------------------------------------------------
void DEVCONN_ICACHE_FLASH
supla_micros() {
	
	if (pomiar != 0) {
		if (nap_prad < 2 )	stan = GPIO_INPUT_GET(supla_w13_pin); 			// pomiar napicia / prdu
		else 				stan = GPIO_INPUT_GET(supla_w14_pin); 			// pomiar mocy
		if (stan == 1) {
			if (stan == stan_poprzedni) {
				microseconds++;							// trwa stan wysoki
				if (microseconds > 40000) {
					voltage_time = 0;
					pomiar = 0;
					microseconds = 0;
					stan_poprzedni = 1;
					od_ready = 1;
				}
			}
			else {										// zbocze narastajce
				if (licznik_pomiarow > 0) {
					if (licznik_pomiarow < 201) voltage_time = voltage_time + microseconds;
					licznik_pomiarow--;
				}
				else {
					voltage_time = voltage_time / 200;
					pomiar = 0;
					od_ready = 1;
				}
				microseconds = 0;
				stan_poprzedni = 1;
			}
		}
		else {
			if (stan == stan_poprzedni) {
				microseconds++; // trwa stan niski
				if (microseconds > 40000) {
					voltage_time = 0;
					pomiar = 0;
					microseconds = 0;
					stan_poprzedni = 0;
					od_ready = 1;
				}
			}
			else {										// zbocze opadajce
				if (licznik_pomiarow > 0) {
					if (licznik_pomiarow < 201) voltage_time = voltage_time + microseconds;
					licznik_pomiarow--;
				}
				else {
					voltage_time = voltage_time / 200;
					pomiar = 0;
					od_ready = 1;
				}
				microseconds = 0;
				stan_poprzedni = 0;
			}
		}
	}
	
}

//-------------------------------------------------------
// Pomiar mocy
//-------------------------------------------------------
void ICACHE_FLASH_ATTR
cf_interrupt() {
    //unsigned long now = supla_micros();
    //hlw8012._last_cf_interrupt = now;
}

//-------------------------------------------------------
// Pomiar napicia i prdu
//-------------------------------------------------------
void ICACHE_FLASH_ATTR
cf1_interrupt() {

   uint32 gpio_status;
   
   gpio_status = GPIO_REG_READ(GPIO_STATUS_ADDRESS);
   GPIO_REG_WRITE(GPIO_STATUS_W1TC_ADDRESS, gpio_status);

}

//-------------------------------------------------------
void setCurrentMultiplier(unsigned long current_multiplier_set) {
	current_multiplier = current_multiplier_set; 
};

void setVoltageMultiplier(unsigned long voltage_multiplier_set) {
	voltage_multiplier = voltage_multiplier_set;
};

void setPowerMultiplier(unsigned long power_multiplier_set) {
	power_multiplier = power_multiplier_set;
};

//-------------------------------------------------------
void ICACHE_FLASH_ATTR
sel_pin_current() {
	supla_esp_gpio_set_hi(SEL_PIN, 1);
}

//-------------------------------------------------------
void ICACHE_FLASH_ATTR
sel_pin_voltage() {
	supla_esp_gpio_set_hi(SEL_PIN, 0);
}

//-------------------------------------------------------
void ICACHE_FLASH_ATTR
setInterrupts() {
	
	//--------------------------------------------------
	// PRD i NAPICIE    15mA / 0.5V na 1Hz
	//--------------------------------------------------
	gpio_output_set(0, 0, 0, GPIO_ID_PIN(13));           //Set GPIO13 as input
	gpio_register_set(GPIO_PIN_ADDR(13), GPIO_PIN_INT_TYPE_SET(GPIO_PIN_INTR_DISABLE)
							| GPIO_PIN_PAD_DRIVER_SET(GPIO_PAD_DRIVER_DISABLE)
							| GPIO_PIN_SOURCE_SET(GPIO_AS_PIN_SOURCE));
	GPIO_REG_WRITE(GPIO_STATUS_W1TC_ADDRESS, BIT(13));   //clear interrupt status
	PIN_PULLUP_EN(PERIPHS_IO_MUX_MTCK_U);			  	 //Enable pin pull up

	//ETS_GPIO_INTR_DISABLE();
    //ETS_GPIO_INTR_ATTACH(cf1_interrupt, NULL );
	//gpio_pin_intr_state_set(GPIO_ID_PIN(13), GPIO_PIN_INTR_ANYEDGE); // Interrupt on any CF1_PIN edge
	//ETS_GPIO_INTR_ENABLE();

	//--------------------------------------------------
	// MOC                12W na 1Hz
	//--------------------------------------------------
	gpio_output_set(0, 0, 0, GPIO_ID_PIN(14));           //Set GPIO14 as input
	gpio_register_set(GPIO_PIN_ADDR(14), GPIO_PIN_INT_TYPE_SET(GPIO_PIN_INTR_DISABLE)
							| GPIO_PIN_PAD_DRIVER_SET(GPIO_PAD_DRIVER_DISABLE)
							| GPIO_PIN_SOURCE_SET(GPIO_AS_PIN_SOURCE));
	GPIO_REG_WRITE(GPIO_STATUS_W1TC_ADDRESS, BIT(14));   //clear interrupt status
	PIN_PULLUP_EN(PERIPHS_IO_MUX_MTMS_U);			  	 //Enable pin pull up

	//--------------------------------------------------
	supla_w13_pin = GPIO_ID_PIN(13);
	supla_w14_pin = GPIO_ID_PIN(14);

}

//-------------------------------------------------------
void ICACHE_FLASH_ATTR
calibrate() {

    // Show corrected factors
    supla_log(LOG_DEBUG, "New current multiplier : ", current_multiplier);
    supla_log(LOG_DEBUG, "New voltage multiplier : ", voltage_multiplier);
    supla_log(LOG_DEBUG, "New power multiplier   : ", power_multiplier);

}

//-------------------------------------------------------
void ICACHE_FLASH_ATTR
supla_pow_setup() {

    hlw8012._voltage = 0;

    hlw8012._last_cf_interrupt = 0;
    hlw8012._last_cf1_interrupt = 0;
    hlw8012._first_cf1_interrupt = 0;

    setCurrentMultiplier(1);
    setVoltageMultiplier(1);
    setPowerMultiplier(1);
	
    supla_log(LOG_DEBUG, "Current multiplier : %u", current_multiplier);
    supla_log(LOG_DEBUG, "Voltage multiplier : %u", voltage_multiplier);
    supla_log(LOG_DEBUG, "Power multiplier   : %u", power_multiplier);
    setInterrupts();

}

//-------------------------------------------------------
void ICACHE_FLASH_ATTR
supla_getVoltage(char value[SUPLA_CHANNELVALUE_SIZE]) {

	hlw8012._voltage = (voltage_time > 0) ? voltage_time : 0;
	memcpy(value, &hlw8012._voltage, sizeof(uint32_t));
}

//-------------------------------------------------------
//void DHT_ICACHE_FLASH supla_pow_init(void) {
//	supla_w1_init();
//}

#endif
